#!  /usr/bin/env python3.7
# -*- coding: utf-8 -*-
from collections import defaultdict, namedtuple
from lztools.text import create_line, center_on, pad_length

Row = namedtuple("Row", ["text", "marks"])

class ColumnWriter(object):
    columns = None
    rows = None
    tests = None
    time = 0
    width = -1
    fail_char = "~"
    bufferd = False
    buffer = None

    def write_line(self, text=""):
        self.buffer.append(text)

    def __init__(self, width=80, fail_char="_"):
        self.rows = []
        self.buffer = []
        self.columns = []
        self.width = width
        self.fail_char = fail_char

    def mark_column(self, text, column, mark):
        if column not in self.columns:
            self.columns.append(column)
        row = None
        for r in self.rows:
            if r.text == text:
                row = r
        if row is None:
            row = Row(text, {column: mark})
            self.rows.append(row)
        if mark not in row.marks:
            row.marks[column] = mark

    def get_column_header(self):
        column_header = "[{}]".format("] [".join(self.columns))
        return pad_length(column_header, self.width, ">")

    def get_statistics(self, text_column_width):
        res = ""
        show_statistics = len(self.columns) > 0
        if show_statistics:
            count = len(self.rows)
            ccounts = defaultdict(int)
            for row in self.rows:
                for column in self.columns:
                    if column in row.marks:
                        if row.marks[column] != self.fail_char:
                            ccounts[column] += 1

            if len(self.columns) > 0:
                res += u"{}".format(create_line(" ", text_column_width, "Statistics:"))
            for c in self.columns:
                text = u"{:.2f}%".format((float(100)/count) * ccounts[c])
                res += center_on(text, u"[{}]".format(c)) + u" "
        return show_statistics, res[0:self.width]

    def _text_width(self):
        column_header = ""
        for c in self.columns:
            column_header += u"[{}] ".format(c)
        column_header = column_header.rstrip()
        return self.width - len(column_header)

    def _create_columns(self):
        ch = self.get_column_header()
        t_width = self._text_width()
        if ch.strip() != "":
            self.write_line(ch)
        for row in self.rows:
            l = create_line(".", t_width, row.text)
            for column in self.columns:
                if column in row.marks:
                    l += center_on(row.marks[column], u"[{}]".format(column)) + " "
                else:
                    l += center_on(self.fail_char, u"[{}]".format(column)) + " "
            self.write_line(l[0:self.width])
        enable_stats, stats = self.get_statistics(t_width)
        if enable_stats:
            self.write_line(stats)

    def flush(self):
        if len(self.columns) > 0:
            self._create_columns()
            o = "\n".join(self.buffer)
            self.buffer = []
            return o



from collections import namedtuple

MatchType = namedtuple("MatchType", ["open", "close"])
brace = MatchType("{", "}")
bracket = MatchType("[", "]")
parentheses = MatchType("(", ")")
gt_lt = MatchType("<", ">")


import inspect
import random
import re
import time
from typing import Union, Iterable, Collection, Container
from ansiwrap import wrap

from lztools import Ansi
from lztools.MatchType import MatchType, brace, bracket, parentheses, gt_lt

_words = None

def words():
    from lztools.bash import command_result
    global _words
    if _words is None:
        _words = command_result("cat", "/usr/share/dict/words")
    return _words

def _get_alignment(alignment:str) -> str:
    if alignment in ["<", "l", "left"]:
        return "<"
    elif alignment in [">", "r", "right"]:
        return ">"
    elif alignment in ["^", "c", "center"]:
        return "^"
    else:
        raise ValueError("Alignment argument not understood")
def _get_padding(padding:int, char:str=" ") -> str:
    result = ""
    for _ in range(0, padding, len(char)):
        result += char
    return result

def create_line(char:str= "-", width:int=200, text:str= "") -> str:
    o = pad_length(text, width, text_alignment="<", pad_char=char)
    return o

def center_on(value:str, text:str) -> str:
    return u"{:^{}}".format(value, len(text))

def pad_length(text:str, width:int, text_alignment:str, pad_char=" ") -> str:
    alignment = _get_alignment(text_alignment)
    if alignment == "^":
        while Ansi.true_length(text) < width:
            text += pad_char
            if Ansi.true_length(text) < width:
                text = pad_char + text
    elif alignment == "<":
        while Ansi.true_length(text) < width:
            text += pad_char
    elif alignment == ">":
        while Ansi.true_length(text) < width:
            text = pad_char + text
    return text

def wall_text(text:str, width:int=80, wall:str= "|", text_alignment="<", h_padding=2, colorizer=None) -> str:
    pad = _get_padding(h_padding)
    text_alignment = _get_alignment(text_alignment)

    result, adjusted = "", width - len(wall) * 2 - h_padding * 2
    executed = False
    for lt in text.splitlines():
        for line in wrap(lt, width=adjusted):
            if colorizer:
                line = colorizer(line)
            executed = True
            line = pad_length(line, adjusted, text_alignment)
            if line == "":
                line = " "
            result += "{}{}{:{}{}}{}{}\n".format(wall, pad, line, text_alignment, adjusted, pad, wall)
    if not executed:
        result = "{}{}{:{}{}}{}{}\n".format(wall, pad, " ", text_alignment, adjusted, pad, wall)
    return result[:-1]

def box_text(text:str, width:int=80, roof:str= "-", wall:str= "|", text_alignment="<") -> str:
    line = pad_length("", width=width, text_alignment=text_alignment, pad_char=roof)
    walled = wall_text(text, wall=wall, text_alignment=text_alignment)
    return f"{line}\n{walled}\n{line}"


def regex(expr:str, text:str, only_first:bool=False, suppress:bool=False) -> str:
    if not only_first:
        return _regex(expr, text, only_first, suppress)
    else:
        try:
            return _regex(expr, text, only_first, suppress).__next__()
        except Exception as e:
            if not suppress:
                raise

def _regex(expr:str, text:str, only_first:bool=False, suppress:bool=False) -> str:
    gen = (x for x in re.findall(expr, text))
    if only_first:
        if suppress:
            try:
                yield gen.__next__()
            except:
                pass
        else:
            yield gen.__next__()
    else:
        yield from gen

def wrap_lines(text: str, width: int = 80) -> str:
    for line in text.splitlines():
        yield from (line[i:i + width] for i in range(0, len(line), width))


def insert_spaces(name:str, underscore:str="") -> str:
    s, n = u"", name[:-4]
    s = s.replace(u"_", underscore)[:-1]
    n = re.sub(r"(?<=\w)([A-Z])", r" \1", str(n))
    return u"{}{}".format(s, n)

def trim_end(remove:str, the_text:str) -> str:
    while the_text.endswith(remove):
        the_text = the_text[:-len(remove)]
    return the_text

def format_seconds(sec:Union[int, float, str]) -> str:
    return time.strftime('%H:%M:%S', time.gmtime(sec))

def search_words(term, strict=False):
    for word in words():
        if strict:
            if term in word:
                yield word
        else:
            pas = True
            for l in set(term):
                if l not in word:
                    pas = False
            if pas:
                yield word

def get_random_word():
    return random.choice(list(words()))

def _is_escaped(text, index) -> bool:
    def __is_escaped(t, i, v) -> bool:
        if t[i-1] == "\\":
            return __is_escaped(t, i-1, not v)
        else:
            return v
    return __is_escaped(text, index, False)


def find_matching(match_type:Union[MatchType, str], text:str, offset:int=0, raise_error:bool=True, fail_value=-1) -> int:
    if match_type is brace or match_type == "{":
        open, close = brace
    elif match_type is bracket or match_type == "[":
        open, close = bracket
    elif match_type is parentheses or match_type == "(":
        open, close = parentheses
    elif match_type is gt_lt or match_type == "<":
        open, close = gt_lt
    else:
        raise ValueError(f"Argument 'match_type' value '{match_type}' not understood.\n'match_type' must be either {{, [, (, < or one of the values from MatchType.")

    depth = 0
    skipping:bool = False
    for i, c in enumerate(text):
        if i < offset:
            continue
        if not skipping:
            if c == '"' and not _is_escaped(text, i):
                skipping = True
            if c == open:
                depth += 1
            if c == close:
                if depth > 0:
                    depth -= 1
                else:
                    return i
        elif skipping and c == '"' and not _is_escaped(text, i):
            skipping = False
    if raise_error:
        raise LookupError("Closing brace not found")
    return fail_value

def as_literal(*args, **kwargs) -> str:
    f = inspect.currentframe().f_back
    filename = inspect.getfile(f)
    code_line = open(filename).readlines()[f.f_lineno - 1]
    t = f"{as_literal.__name__}("
    i = code_line.find(t) + len(t)
    e = find_matching(parentheses, code_line, offset=i)
    return code_line[i:e]


def print_collection(obj):
    _print_collection(obj)

def _print_collection(ob, i=0):
    indent = "".join(["\t" for _ in range(i)])
    if type(ob) == list or type(ob) == tuple:
        for i, x in enumerate(ob):
            res = f"{indent}{i}:"
            if is_collection(x):
                print(res)
                _print_collection(x, i)
            else:
                print(f"{res} {x}")
    elif type(ob) == dict:
        for key in ob:
            o = ob[key]
            res = f"{indent}{key}:"
            if type(o) in [list, dict, tuple]:
                print(res)
                _print_collection(o, i + 1)
            else:
                print(f"{res} {o}")
    else:
        print(ob)

def print_dict(obj):
    def pd(ob, i=0):
        indent = "".join(["\t" for _ in range(i)])
        for key in ob.__dict__:
            o = ob.__dict__[key]
            res = f"{indent}{key}:"
            if is_collection(o):
                print(res)
                _print_collection(o, i + 1)
            else:
                print(f"{res} {o}")

    pd(obj)

def is_collection(obj):
    return type(obj) in [list, dict, tuple]

def print_dir_values(obj):
    def pd(ob, i=0):
        indent = "".join(["\t" for _ in range(i)])
        for key in dir(ob):
            o = getattr(ob, key)
            res = f"{indent}{key}:"
            if is_collection(o):
                print(res)
                _print_collection(o, i + 1)
            else:
                print(f"{res} {o}")

    pd(obj)



from lztools.text import create_line, wall_text

default_horizontal_symbol = "-"
default_vertical_symbol = "|"
default_fill_symbol = " "

class BlockWriter(object):
    roof_str = None
    wall_str = None
    fill_char = None
    started = False
    buffer = None
    width = -1
    padding_h = -1

    def __init__(self, width:int=80, horizontal_symbol:str=default_horizontal_symbol, vertical_symbol:str=default_vertical_symbol, fill_symbol:str=default_fill_symbol, padding_h:int=2):
        self.roof_str = horizontal_symbol
        self.wall_str = vertical_symbol
        self.fill_char = fill_symbol
        self.width = width
        self.padding_h = padding_h
        self.buffer = []

    def _wall_len(self):
        return len(self.wall_str)

    def split(self) -> None:
        w = self.width - 2 * self._wall_len()
        self.buffer.append(f"{self.wall_str}{create_line(width=w)}{self.wall_str}")

    def seperate(self):
        self.write_text(" ")

    def square_text(self, text:str, text_alignment="^") -> None:
        w = self.width - 2 * self._wall_len()
        if not self.started:
            self.started = True
            self.buffer.append(create_line(self.roof_str, self.width))
        else:
            self.buffer.append(f"{self.wall_str}{create_line(width=w)}{self.wall_str}")
        self.seperate()
        self.write_text(text, text_alignment=text_alignment)
        self.seperate()
        self.buffer.append(f"{self.wall_str}{create_line(width=w)}{self.wall_str}")

    def write_text(self, text:str, text_alignment="<", colorizer=None) -> None:
        if not self.started:
            self.buffer.append(create_line(self.roof_str, self.width))
            self.started = True
        self.buffer.append(wall_text(text, self.width, self.wall_str, text_alignment=text_alignment, h_padding=self.padding_h, colorizer=colorizer))

    def write_output(self, print_func=print) -> None:
        for line in self.buffer:
            print_func(line)
        print_func(create_line(self.roof_str, self.width))


import re

_c_ = {
    "Red": "\033[31m",
    "Cyan": "\033[96m",
    "Yellow": "\033[33m",
    "Bold": "\033[1m",
    "Reverse": "\033[7m",
    "Reset": "\033[0m"
}

__extra_symbols = [
    "\x1b",
    "\x00",
    "\x01",
    "\x02",
    "\x03",
    "\x04",
    "\x05",
    "\x06",
    "\x07",
    "\x08",
    "\x09",
    "\x10"
]

def true_length(string):
    for astr in _c_.values():
        string = string.replace(astr, "")
    return len(string)

def cyan(string):
    return "{}{}{}".format(_c_['Cyan'], string, _c_['Reset'])

def red(string):
    return "{}{}{}".format(_c_['Red'], string, _c_['Reset'])

def yellow(string):
    return "{}{}{}".format(_c_['Yellow'], string, _c_['Reset'])

def bold(string):
    return "{}{}{}".format(_c_['Bold'], string, _c_['Reset'])

def reverse(string):
    return "{}{}{}".format(_c_['Reverse'], string, _c_['Reset'])

def strip(text):
    for c in _c_.values():
        text = text.replace(c, "")
    for s in __extra_symbols:
        text = text.replace(s, "")
    return text

def strip_reg(text):
    text = text.encode('string-escape')
    text = re.sub(r"\\x1b", '', text)
    text = re.sub(r"\\x[0-9][0-9]", '', text)
    return text.decode('string-escape')

def ansi_length(string, count_close=True):
    rst = _c_["Reset"]
    lis = _c_.values()
    if not count_close:
        lis.remove(rst)
    for astr in lis:
        string = string.replace(astr, " ")
    string = string.replace(rst, "")
    return len(string)

# def ansi_length(string):
#     for astr in _c_.values():
#         string = string.replace(astr, " ")
#     return len(string)
